#include <Wire.h>
#include <SoftwareSerial.h>
#define TX_PIN               0
#define RX_PIN               1
#define BAUDRATE             115200

#define MPU925_ADD  0x68
#define AK8963_ADD  0x0C
#define WAI_MP   0x75
#define MP_WAKE  0x6B
#define MAG_WAKE 0x37
#define MAG_AD_ADD 0x0A
#define MAG_AD_ST 0x12

#define ACC_X_H 0x3B
#define ACC_X_L 0x3C
#define ACC_Y_H 0x3D
#define ACC_Y_L 0x3E
#define ACC_Z_H 0x3F
#define ACC_Z_L 0x40

#define ACC_X_ST 0x0D
#define ACC_Y_ST 0x0E
#define ACC_Z_ST 0x0F

#define TMP_H 0x41
#define TMP_L 0x42

#define GYR_X_H 0x43
#define GYR_X_L 0x44
#define GYR_Y_H 0x45
#define GYR_Y_L 0x46
#define GYR_Z_H 0x47
#define GYR_Z_L 0x48

#define GYR_X_ST 0x00
#define GYR_Y_ST 0x01
#define GYR_Z_ST 0x02


#define MAG_X_H 0x03
#define MAG_X_L 0x04
#define MAG_Y_H 0x05
#define MAG_Y_L 0x06
#define MAG_Z_H 0x07
#define MAG_Z_L 0x08
#define MAG_16 0x10

#define TMP_OFF -6.0

int ACC_X_D;
int ACC_Y_D;
int ACC_Z_D;
int GYR_X_D;
int GYR_Y_D;
int GYR_Z_D;
int MAG_X_D;
int MAG_Y_D;
int MAG_Z_D;
int TMP_D;
int cnt=0;

float ACC_X;
float ACC_Y;
float ACC_Z;
float GYR_X;
float GYR_Y;
float GYR_Z;
float MAG_X;
float MAG_Y;
float MAG_Z;
float TMP_;

float ACC_ST[4];
float GYR_ST[4];
float MAG_ST[4];
float ACC_OFF[4];
float GYR_OFF[4];
float MAG_OFF[4];

int16_t ACC_(uint8_t ADD,uint8_t H_ADD,uint8_t L_ADD);
SoftwareSerial FioSerial = SoftwareSerial( RX_PIN, TX_PIN );

/////////////////////////////////////////////////

/////////////////////////////////////////////////
void write_i2C(uint8_t DEV_ADD,uint8_t Reg_ADD,uint8_t VAL){ //RasPI風
   
  Wire.beginTransmission(DEV_ADD);
  Wire.write(Reg_ADD);
  Wire.write(VAL);
  Wire.endTransmission();
 
  
  
}

/////////////////////////////////////////////////

/////////////////////////////////////////////////
uint8_t read_i2c(uint8_t DEV_ADD,uint8_t Reg_ADD,uint8_t READ_NUM){ //RasPI風
  Wire.beginTransmission(DEV_ADD);
  Wire.write(Reg_ADD);
  Wire.endTransmission();
  
  Wire.requestFrom(DEV_ADD,READ_NUM);
  while(Wire.available()<1){
  } 
  return Wire.read();
  
  
}

/////////////////////////////////////////////////

/////////////////////////////////////////////////
void initSerial( void ){
  pinMode( TX_PIN, OUTPUT );
  FioSerial.begin(BAUDRATE);
}////////////////////////////////////////////

/////////////////////////////////////////////////
void setup( ){
  initSerial( );
  Serial.begin(BAUDRATE);
  Wire.begin(); // I2Cバスに接続
  delay(10);
  Wire.beginTransmission(MPU925_ADD);
  Wire.write(WAI_MP);
  Wire.endTransmission();
  
  Wire.requestFrom(MPU925_ADD, 1);
  while(Wire.available()<1){
  } 
  //Serial.print("0x");
  //Serial.print(Wire.read(),HEX);
  //Serial.print("\n\r");
  
  
  Wire.beginTransmission(MPU925_ADD);
  Wire.write(MP_WAKE);
  Wire.write(0x00);
  Wire.endTransmission();
  
  Wire.beginTransmission(MPU925_ADD);
  Wire.write(MAG_WAKE);
  Wire.write(0x02);
  Wire.endTransmission();
   Calib();
  Serial.print("\n\r");
}

/////////////////////////////////////////////////

/////////////////////////////////////////////////
void Calib(){
  int i=0;

  GYR_OFF[0]= -183.49;
  GYR_OFF[1]= 44.00;
  GYR_OFF[2]= 53.87;
  
  ACC_OFF[0] = 35.44;
  ACC_OFF[1] = -225.52;
  ACC_OFF[2] = 695.78;


  
}
/////////////////////////////////////////////////

/////////////////////////////////////////////////
int16_t ADJ_(int16_t S_data){
  int16_t bit_;
  if(S_data & (0x01<<15) ){
    bit_ = S_data^0xffff;
    bit_ = bit_+1;
    S_data =- bit_;
  }
  return S_data;
}

/////////////////////////////////////////////////

/////////////////////////////////////////////////
int16_t ACC_(uint8_t ADD,uint8_t H_ADD,uint8_t L_ADD){
  
  int16_t DAT=0;
  int DAT_H;
  int DAT_L;
  int DAT_ADJ;
  signed int Signed;
  unsigned int Unsigned;
 
if(ADD==MPU925_ADD){
  Wire.beginTransmission(ADD);
  Wire.write(H_ADD);
  Wire.endTransmission();
  
  Wire.requestFrom(ADD,(uint8_t)8);
  while(Wire.available()<1){
  } 
  DAT_H=Wire.read();
  Wire.beginTransmission(ADD);
  Wire.write(L_ADD);
  Wire.endTransmission();
  
  Wire.requestFrom(ADD,(uint8_t)8);
  while(Wire.available()<1){
  } 
  DAT_L=Wire.read();
  DAT =(int16_t)( ((int16_t)DAT_H <<8 )| DAT_L ); 
}else{
  Wire.beginTransmission(ADD);
  Wire.write(L_ADD);
  Wire.endTransmission();
  
  Wire.requestFrom(ADD,(uint8_t)8);
  while(Wire.available()<1){
  } 
  DAT_H=Wire.read();
  //Serial.print("ACC_X_H:0x");
  //Serial.print(DAT_H,HEX);
 // Serial.print("\n\r");
  
  Wire.beginTransmission(ADD);
  Wire.write(H_ADD);
  Wire.endTransmission();
  
  Wire.requestFrom(ADD,(uint8_t)8);
  while(Wire.available()<1){
  } 
  DAT_L=Wire.read();
  DAT =(int16_t)( ((int16_t)DAT_H <<8 )| DAT_L ); 
  
  
  
}
if(ADD==MPU925_ADD){
   DAT_ADJ=ADJ_(DAT);
}else{
   DAT_ADJ= (signed int)DAT;
  //unsigned2signed(DAT);

}

   return DAT_ADJ;
}

/////////////////////////////////////////////////

/////////////////////////////////////////////////
void loop ( ){
      int sensorVal[5];
      int i=0;
      int STB;
      int COMP_D; 
      double mRes = 10.*4912./32760.0; 
    //  Serial.print("N");
      ACC_X_D=ACC_(MPU925_ADD,ACC_X_H,ACC_X_L);
     // Serial.print("ACC_X:");
      ACC_X=(float)(ACC_X_D-ACC_OFF[0])/16384;
    //  Serial.print(ACC_X);
 //     Serial.print("\n\r");

      ACC_Y_D=ACC_(MPU925_ADD,ACC_Y_H,ACC_Y_L);
  //    Serial.print("ACC_Y:");
      ACC_Y=(float)(ACC_Y_D-ACC_OFF[1])/16384;
  //    Serial.print(ACC_Y);
   //   Serial.print("\n\r");     
      
      
      
      ACC_Z_D=ACC_(MPU925_ADD,ACC_Z_H,ACC_Z_L);
   //   Serial.print("ACC_Z:");
      ACC_Z=(float)(ACC_Z_D-ACC_OFF[2])/16384;
   //   Serial.print(ACC_Z);
   //   Serial.print("\n\r");
      
      
      
      GYR_X_D=ACC_(MPU925_ADD,GYR_X_H,GYR_X_L);
  //    Serial.print("GYR_X:");
        GYR_X=(float)(GYR_X_D-GYR_OFF[0])/131;
  //    Serial.print(GYR_X);
  //    Serial.print("\n\r");
      
      GYR_Y_D=ACC_(MPU925_ADD,GYR_Y_H,GYR_Y_L);
    //  Serial.print("GYR_Y:");
      GYR_Y=(float)(GYR_Y_D-GYR_OFF[1])/131;
   //   Serial.print(GYR_Y);
   //   Serial.print("\n\r");           
      GYR_Z_D=ACC_(MPU925_ADD,GYR_Z_H,GYR_Z_L);
      
      //Serial.print("GYR_Z:");
      GYR_Z=((float)(GYR_Z_D-GYR_OFF[2])/131);
      //Serial.print(GYR_Z);
      //Serial.print("\n\r"); 
     // Serial.print(cnt);
    //  Serial.print("\n\r");
    
      TMP_D=ACC_(MPU925_ADD,TMP_H,TMP_L);
    //  Serial.print("GYR_Y:");
      TMP_=(float)(TMP_D)/333.87+21+TMP_OFF;
      Wire.beginTransmission(AK8963_ADD);
      Wire.write(MAG_AD_ADD);
      Wire.write(MAG_AD_ST);
      Wire.endTransmission();
      
      
      MAG_X_D= ADJ_( ACC_(AK8963_ADD ,MAG_X_H,MAG_X_L) );
     // Serial.print("MAG_X:");
        MAG_X=(float)(MAG_X_D)*0.15*mRes;
     // Serial.print((int16_t)MAG_X);
     // Serial.print("\n\r");
      
      MAG_Y_D= ADJ_( ACC_(AK8963_ADD ,MAG_Y_H,MAG_Y_L) );
     // Serial.print("MAG_Y:");
        MAG_Y=(float)(MAG_Y_D)*0.15*mRes;
     // Serial.print(MAG_Y);
     // Serial.print("\n\r");
      
      MAG_Z_D= ADJ_( ACC_(AK8963_ADD ,MAG_Z_H,MAG_Z_L) );
     // Serial.print("MAG_Z:");
        MAG_Z=(float)(MAG_Z_D)*0.15*mRes;
     // Serial.print(MAG_Z);
     // Serial.print("\n\r");
     // */
     
     
     
     
     
     
     char rcv;
     int m_cnt=0;
    m_cnt=Serial.available();
     if (m_cnt) {
		rcv = Serial.read();
                if(rcv == 'R' ){ 
                        Serial.print(ACC_X);  
                        Serial.print(":");
                        Serial.print(ACC_Y);  
                        Serial.print(":");
                        Serial.print(ACC_Z);  
                        Serial.print(":");
                        Serial.print(GYR_X);  
                        Serial.print(":");
                        Serial.print(GYR_Y);  
                        Serial.print(":");
                        Serial.print(GYR_Z);  
                        Serial.print(":");
                        Serial.print(MAG_X);  
                        Serial.print(":");  
                        Serial.print(MAG_Y);    
                        Serial.print(":");
                        Serial.print(MAG_Z);  
                        Serial.print(":");
                        Serial.println(TMP_); 
                }else if(rcv == 'C' ){
                Calib();
                }
      //      Serial.print(":");
//     Serial.print(cnt);
     }else{
     }
 //    Serial.flush();
     delay(1);
     /* 
     
//    cnt++;
      /*
       Serial.println(sensorVal[0]);
       Serial.println(sensorVal[1]);
       Serial.println(sensorVal[2]);
       Serial.println(sensorVal[3]);
       */
 
}


