// -*- C++ -*-
/*!
 * @file  RTC_NineDOF_Ardiuno.cpp
 * @brief 9軸センサをつないだArdiuno（または何らかのマイコン）からセンサデータ取得。機体位置姿勢推定
 * @date $Date$
 *
 * @author AIZUK
 * kohno@aizuk.jp
 *
 * $Id$
 */

#include "RTC_NineDOF_Ardiuno.h"

#define _DIS(x,y)	(sqrt( (x)*(x) + (y)*(y) ))
#define degree		(180 / M_PI)
#define rad		(M_PI / 180)
/************************************************************************************
/			変数定義						　　　 /
/			                    					    /
************************************************************************************/
//////////////////////////////////////////////
//     Ardiuno
//////////////////////////////////////////////
int Ardiuno;
int SERIAL_OK_FLG[5];
struct termios Ardiuno_ter;
//////////////////////////////////////////////
//	Sensor Data
//////////////////////////////////////////////
double ACC_Raw[3]={0,0,0};//加速度　　生データ
double GYR_Raw[3]={0,0,0};//角速度　　生データ
double MAG_Raw[3]={0,0,0};//磁束強度　生データ
double TMP_Raw=0;//温度　生データ

double ACC_R[3]={0,0,0};//加速度　　座標変換データ
double G_Global[3]={0,0,1};//加速度　　座標変換データ
double G_R[3]={0,0,0};//加速度　　座標変換データ
double GYR_R[3]={0,0,0};//角速度　　座標変換データ
double MAG_R[3]={0,0,0};//磁束強度　座標変換データ

double ACC_Med[3]={0,0,0};//加速度　　メディアンフィルタデータ
double ACC_Med_OLD_X[13]={0,0,0,0,0,0,0,0,0,0,0,0,0};//加速度　　メディアンフィルタデータ
double ACC_Med_OLD_Y[13]={0,0,0,0,0,0,0,0,0,0,0,0,0};//加速度　　メディアンフィルタデータ
double ACC_Med_OLD_Z[13]={0,0,0,0,0,0,0,0,0,0,0,0,0};//加速度　　メディアンフィルタデータ
double ACC_Med_Gcan[3]={0,0,0};//加速度　　メディアンフィルタデータ
double ACC_Med_Gcan_OLD_X[13]={0,0,0,0,0,0,0,0,0,0,0,0,0};//加速度　　メディアンフィルタデータ
double ACC_Med_Gcan_OLD_Y[13]={0,0,0,0,0,0,0,0,0,0,0,0,0};//加速度　　メディアンフィルタデータ
double ACC_Med_Gcan_OLD_Z[13]={0,0,0,0,0,0,0,0,0,0,0,0,0};//加速度　　メディアンフィルタデータ
double GYR_Med[3]={0,0,0};//角速度　　メディアンフィルタデータ
double GYR_Med_OLD_X[13]={0,0,0,0,0,0,0,0,0,0,0,0,0};//角速度　　メディアンフィルタデータ
double GYR_Med_OLD_Y[13]={0,0,0,0,0,0,0,0,0,0,0,0,0};//角速度　　メディアンフィルタデータ
double GYR_Med_OLD_Z[13]={0,0,0,0,0,0,0,0,0,0,0,0,0};//角速度　　メディアンフィルタデータ
double MAG_Med[3]={0,0,0};//磁束強度　メディアンフィルタデータ;
double MAG_Med2[3]={0,0,0};//磁束強度　メディアンフィルタデータ;
double MAG_OFF[3]={0,0,0};//磁束強度　メディアンフィルタデータ;
double MAG_Med_OLD_X[13]={0,0,0,0,0,0,0,0,0,0,0,0,0};//磁束強度　　メディアンフィルタデータ
double MAG_Med_OLD_Y[13]={0,0,0,0,0,0,0,0,0,0,0,0,0};//磁束強度　　メディアンフィルタデータ
double MAG_Med_OLD_Z[13]={0,0,0,0,0,0,0,0,0,0,0,0,0};//磁束強度　メディアンフィルタデータ
double TMP_Med=0;//温度　生データ
double TMP_Med_OLD[13]={0,0,0,0,0,0,0,0,0,0,0,0,0};//磁束強度　　メディアンフィルタデータ
//////////////////////////////////////////////
//	Orientation
//////////////////////////////////////////////
double Direction;
double Direction_OLD;
double roll;
double pitch;
double yaw;
double roll_A; //roll acc
double pitch_A;//pitch acc
double yaw_M;  //yaw mag
double roll_IN; //roll Inport
double pitch_IN;//pitch Inport
double yaw_IN;  //yaw Inport
double roll_dt; //roll Inport
double pitch_dt;//pitch Inport
double yaw_dt;  //yaw Inport
int Dir_cnt=0;
int Ori_dt_Flg=0;
int MAG_Turn_Flg[5]={0,0,0,0,0};
double MAG_Turn_data[13]={0,0,0,0,0,0,0,0,0,0,0,0,0};//磁束強度　メディアンフィルタデータ
double MAG_Turn_def[13]={0,0,0,0,0,0,0,0,0,0,0,0,0};//磁束強度　メディアンフィルタデータ
double MAG_Turn_def2[13]={0,0,0,0,0,0,0,0,0,0,0,0,0};//磁束強度　メディアンフィルタデータ
double MAG_Med_T_OLD[13]={0,0,0,0,0,0,0,0,0,0,0,0,0};//磁束強度　メディアンフィルタデータ
int MAG_Med_T_FLG[13]={1,1,1,1,1,1,1,1,1,1,1,1,1};//磁束強度　メディアンフィルタデータ
int MAG_Med_T_FLG2[13]={0,0,0,0,0,0,0,0,0,0,0,0,0};//磁束強度　メディアンフィルタデータ
//////////////////////////////////////////////
//	pOSE
//////////////////////////////////////////////
double Velocity_[4]={0,0,0,0};
//////////////////////////////////////////////
//	Kalman
//////////////////////////////////////////////
double YPR[4];//0:yaw 1:pitch 2:roll
double YPR_O[4];
double YPR_[4];
double P_YPR_[3][3]={{1 ,0 ,0},
	 	 {0 ,1 ,0},  
 		 {0 ,0 ,1} };
double P_YPR[3][3]={{1 ,0 ,0},
	 	 {0 ,0.1 ,0},  
 		 {0 ,0 ,1} };

double K[3][3];
int Orie_In_Flg=0;
int Pose_In_Flg=0;
int Posi_In_Flg=0;
int Orie_In_Flg_Once=0;
int Pose_In_Flg_Once=0;
int Posi_In_Flg_Once=0;

double XYZ[4];//0:yaw 1:pitch 2:roll
double XYZ_dt[4];//0:yaw 1:pitch 2:roll
double XYZ_O[4];
double XYZ_[4];
double P_XYZ_[3][3]={{1 ,0 ,0},
	 	 {0 ,1 ,0},  
 		 {0 ,0 ,1} };
double P_XYZ[3][3]={{1 ,0 ,0},
	 	 {0 ,0.1 ,0},  
 		 {0 ,0 ,1} };
double K_XYZ[3][3];
//////////////////////////////////////////////
//	Time
//////////////////////////////////////////////
double t0=0,t1=0;
double dt=0,sum_t=0;
struct timeval ti;
int First_Flg=1;
int cnt=0;


using namespace std;
ofstream Logout_V( "NineDOF_LOGS_SensingData.csv", ios::trunc );
ofstream Logout_K( "NineDOF_LOGS_Pose.csv", ios::trunc );
double Direction_Estimate(double ACC_Med[3],double MAG_Med[3],double r,double p,double OFFSET);
double Median(int Len,double new_data,double* data);
void cal_YPR(double* ACC_,double* GYR_,double* MAG_,double Y_0,double Y_1,double Y_2);
void cal_XYZ(double* ACC_,double X,double Y,double Z);
void LOG_init();

// Module specification
// <rtc-template block="module_spec">
static const char* rtc_ninedof_ardiuno_spec[] =
  {
    "implementation_id", "RTC_NineDOF_Ardiuno",
    "type_name",         "RTC_NineDOF_Ardiuno",
    "description",       "9軸センサをつないだArdiuno（または何らかのマイコン）からセンサデータ取得。機体位置姿勢推定",
    "version",           "1.0.0",
    "vendor",            "AIZUK",
    "category",          "Sensor",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
    "conf.default.Kalman_ON_OFF", "1",
    "conf.default.Port", "/dev/ttyUSB3",
    "conf.default.BAUDRATE", "B115200",
    "conf.default.Median_ON_OFF", "1",
    "conf.default.Orientation_FIX", "0",
    "conf.default.Position_FIX", "0",
    "conf.default.Median_NUM", "3",
    "conf.default.Yaw_OFFSET", "0.0",
    "conf.default.Pitch_OFFSET", "-0.42",
    "conf.default.Roll_OFFSET", "-1.72",
    // Widget
    "conf.__widget__.Kalman_ON_OFF", "radio",
    "conf.__widget__.Port", "text",
    "conf.__widget__.BAUDRATE", "radio",
    "conf.__widget__.Median_ON_OFF", "radio",
    "conf.__widget__.Orientation_FIX", "radio",
    "conf.__widget__.Position_FIX", "radio",
    "conf.__widget__.Median_NUM", "text",
    "conf.__widget__.Yaw_OFFSET", "text",
    "conf.__widget__.Pitch_OFFSET", "text",
    "conf.__widget__.Roll_OFFSET", "text",
    // Constraints
    "conf.__constraints__.Kalman_ON_OFF", "(0,1)",
    "conf.__constraints__.BAUDRATE", "(B9600,B14400,B19200,B38400,B57600,B115200)",
    "conf.__constraints__.Median_ON_OFF", "(0,1)",
    "conf.__constraints__.Orientation_FIX", "(0,1,2,3,4,5,6,7)",
    "conf.__constraints__.Position_FIX", "(0,1,2,3,4,5,6,7)",
    ""
  };
// </rtc-template>

/*!
 * @brief constructor
 * @param manager Maneger Object
 */
RTC_NineDOF_Ardiuno::RTC_NineDOF_Ardiuno(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_Pose_InIn("Pose_In", m_Pose_In),
    m_Position_InIn("Position_In", m_Position_In),
    m_Orientation_InIn("Orientation_In", m_Orientation_In),
    m_Pose_OutOut("Pose_Out", m_Pose_Out),
    m_Position_OutOut("Position_Out", m_Position_Out),
    m_Orientation_OutOut("Orientation_Out", m_Orientation_Out),
    m_AccOut("Acc", m_Acc),
    m_GyroOut("Gyro", m_Gyro),
    m_MagOut("Mag", m_Mag),
    m_TmpOut("Tmp", m_Tmp)

    // </rtc-template>
{
}

/*!
 * @brief destructor
 */
RTC_NineDOF_Ardiuno::~RTC_NineDOF_Ardiuno()
{
}



RTC::ReturnCode_t RTC_NineDOF_Ardiuno::onInitialize()
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  addInPort("Pose_In", m_Pose_InIn);
  addInPort("Position_In", m_Position_InIn);
  addInPort("Orientation_In", m_Orientation_InIn);
  
  // Set OutPort buffer
  addOutPort("Pose_Out", m_Pose_OutOut);
  addOutPort("Position_Out", m_Position_OutOut);
  addOutPort("Orientation_Out", m_Orientation_OutOut);
  addOutPort("Acc", m_AccOut);
  addOutPort("Gyro", m_GyroOut);
  addOutPort("Mag", m_MagOut);
  addOutPort("Tmp", m_TmpOut);
  
  // Set service provider to Ports
  
  // Set service consumers to Ports
  
  // Set CORBA Service Ports
  
  // </rtc-template>

  // <rtc-template block="bind_config">
  // Bind variables and configuration variable
  bindParameter("Kalman_ON_OFF", m_Kalman_ON_OFF, "1");
  bindParameter("Port", m_Port, "/dev/ttyUSB3");
  bindParameter("BAUDRATE", m_BAUDRATE, "B115200");
  bindParameter("Median_ON_OFF", m_Median_ON_OFF, "1");
  bindParameter("Orientation_FIX", m_Orientation_FIX, "0");
  bindParameter("Position_FIX", m_Position_FIX, "0");
  bindParameter("Median_NUM", m_Median_NUM, "3");
  bindParameter("Yaw_OFFSET", m_Yaw_OFFSET, "0.0");
  bindParameter("Pitch_OFFSET", m_Pitch_OFFSET, "-0.42");
  bindParameter("Roll_OFFSET", m_Roll_OFFSET, "-1.72");
  // </rtc-template>
  sum_t=0;
  cnt-0;
  MAG_Med_T_FLG[0] = 1;
  MAG_Med_T_FLG[1] = 1;
  MAG_Med_T_FLG[2] = 1;

  
  LOG_init();
  return RTC::RTC_OK;
}


RTC::ReturnCode_t RTC_NineDOF_Ardiuno::onFinalize()
{
  return RTC::RTC_OK;
}


RTC::ReturnCode_t RTC_NineDOF_Ardiuno::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}


RTC::ReturnCode_t RTC_NineDOF_Ardiuno::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}


RTC::ReturnCode_t RTC_NineDOF_Ardiuno::onActivated(RTC::UniqueId ec_id)
{
  Connect_serial();//マイコンとシリアル通信・接続
  usleep(51000); 
  return RTC::RTC_OK;
}


RTC::ReturnCode_t RTC_NineDOF_Ardiuno::onDeactivated(RTC::UniqueId ec_id)
{
  close(Ardiuno);//マイコンとシリアル通信・接続終了
  return RTC::RTC_OK;
}


RTC::ReturnCode_t RTC_NineDOF_Ardiuno::onExecute(RTC::UniqueId ec_id)
{
int i,j,k;
	int red;
	double tmp=0;
	char def[1];
	char buf[512]="";
///////////////////////////////////////////////////////////////////////////////////////////////////////////
//    マイコンデータ読み出し 	
///////////////////////////////////////////////////////////////////////////////////////////////////////////

//よみだし命令
	sprintf(buf,"R");
	for(i=0;i<strlen(buf);i++){
		def[0]=buf[i];
		red=write(Ardiuno,def,sizeof(def));
		if (red == -1) {
			printf("Write_ERROR\n\r");
		}
	}
	
	usleep(51000); 
	memset(&buf, 0, sizeof(buf));
//よみだし
	red=read(Ardiuno, buf,512);
	if (red == -1) {
		printf("Read_ERROR\n\r");
	}else{
	//	printf("www\n\r");
		printf("%s\n",buf);
		sscanf(buf,"%lf:%lf:%lf:%lf:%lf:%lf:%lf:%lf:%lf:%lf",&ACC_Raw[0],&ACC_Raw[1],&ACC_Raw[2],
		&GYR_Raw[0],&GYR_Raw[1],&GYR_Raw[2],&MAG_Raw[0],&MAG_Raw[1],&MAG_Raw[2],&TMP_Raw);
	}
	
///////////////////////////////////////////////////////////////////////////////////////////////////////////
//     メディアンフィルタ 	スパイクノイズ除去
///////////////////////////////////////////////////////////////////////////////////////////////////////////

	if(m_Median_ON_OFF==1){		//メディアンフィルタフラグ
	
		//生データにメディアンフィルタ　以下_Medはメディアンフィルタ後のデータ
		ACC_Med[0]=Median(m_Median_NUM,ACC_Raw[0],ACC_Med_OLD_X);	//ACC_X
		ACC_Med[1]=Median(m_Median_NUM,ACC_Raw[1],ACC_Med_OLD_Y);	//ACC_Y
		ACC_Med[2]=Median(m_Median_NUM,ACC_Raw[2],ACC_Med_OLD_Z);	//ACC_Z
		GYR_Med[0]=Median(m_Median_NUM,GYR_Raw[0],GYR_Med_OLD_X);	//GYR_X
		GYR_Med[1]=Median(m_Median_NUM,GYR_Raw[1],GYR_Med_OLD_Y);	//GYR_Y
		GYR_Med[2]=Median(m_Median_NUM,GYR_Raw[2],GYR_Med_OLD_Z);	//GYR_Z
		MAG_Med[0]=Median(m_Median_NUM,MAG_Raw[0],MAG_Med_OLD_X);	//MAG_X
		MAG_Med[1]=Median(m_Median_NUM,MAG_Raw[1],MAG_Med_OLD_Y);	//MAG_Y
		MAG_Med[2]=Median(m_Median_NUM,MAG_Raw[2],MAG_Med_OLD_Z);	//MAG_Z
		TMP_Med=Median(m_Median_NUM,TMP_Raw,TMP_Med_OLD);	//TMP
		
		

	}else{		//メディアンフィルタフラグなし


		//生データそのまま　以下_Medは生データ
		ACC_Med[0]=ACC_Raw[0];
		ACC_Med[1]=ACC_Raw[1];
		ACC_Med[2]=ACC_Raw[2];
		GYR_Med[0]=GYR_Raw[0];
		GYR_Med[1]=GYR_Raw[1];
		GYR_Med[2]=GYR_Raw[2];
		MAG_Med[0]=MAG_Raw[0];
		MAG_Med[1]=MAG_Raw[1];
		MAG_Med[2]=MAG_Raw[2];
		TMP_Med=TMP_Raw;
		

	}
///////////////////////////////////////////////////////////////////////////////////////////////////////////
//     地磁気　重力補正　 yaw算出
///////////////////////////////////////////////////////////////////////////////////////////////////////////
	if(cnt>=m_Median_NUM){
		if(ACC_Med[2]>0){
				 	roll_A=atan(ACC_Med[1]/ACC_Med[2])+ (m_Roll_OFFSET)*rad;
			 	}else if(ACC_Med[2]==0){
			 	 	roll_A=0+ (m_Roll_OFFSET)*rad;
			 	}else{
			 	 	roll_A=atan(ACC_Med[1]/ACC_Med[2])+M_PI/2+ (m_Roll_OFFSET)*rad;
			  	}
		tmp=ACC_Med[1]*ACC_Med[1]+ACC_Med[2]*ACC_Med[2];
		pitch_A=atan2((-ACC_Med[0]),(sqrt(tmp)))+ (m_Pitch_OFFSET)*rad;
		
		
		if(MAG_Med_T_FLG[1]==1){
		MAG_OFF[0]=-54.96-22+6.565-1.58-19.34;//42.28;
		MAG_OFF[1]=54.88+5.96+1.275-62.34;
		MAG_OFF[2]=21.255;//-50.575;//MAG_Med[2];
		MAG_Med_T_FLG[1]=0;
		}
		MAG_Med[0]=MAG_Med[0]-MAG_OFF[0];
		MAG_Med[1]=MAG_Med[1]-MAG_OFF[1];
		MAG_Med[2]=MAG_Med[2]-MAG_OFF[2];
		
		MAG_Med_T_OLD[0]=MAG_Med[0];
		MAG_Med_T_OLD[1]=MAG_Med[1];
		MAG_Med_T_OLD[2]=MAG_Med[2];
		printf("med_flg:%f \n\r",MAG_Med_T_OLD[1]-MAG_Med[1]);
		printf("med_flg:%d \n\r",MAG_Med_T_FLG[1]);
 		
		MAG_Med[0]=MAG_Med[0];//23.615;
		MAG_Med[1]=MAG_Med[1];//22.905;
		MAG_Med[2]=MAG_Med[2]/32.950;
	
		yaw_M =Direction_Estimate(ACC_Med,MAG_Med,roll,pitch,m_Yaw_OFFSET);
		
		if(Ori_dt_Flg==0){
			roll_dt += roll_A + m_Roll_OFFSET;
			pitch_dt += pitch_A + m_Pitch_OFFSET;
			yaw_dt  += yaw_M + m_Yaw_OFFSET;
		}

///////////////////////////////////////////////////////////////////////////////////////////////////////////
//     カルマンフィルタ 	YawPitchRoll
///////////////////////////////////////////////////////////////////////////////////////////////////////////

	Kal_Cal_YPR();
	Ori_dt_Flg=1;
	
	
		
///////////////////////////////////////////////////////////////////////////////////////////////////////////
//     メディアンフィルタ 	G-Cancel
///////////////////////////////////////////////////////////////////////////////////////////////////////////

	if(m_Median_ON_OFF==1){		//メディアンフィルタフラグ

	//ロボット座標系でのACC
		ACC_R[2]= ACC_Med[2]*( cos(roll*rad)*cos(pitch*rad))
	    	       + ACC_Med[1]*(cos(roll*rad)*sin(pitch*rad)*sin(yaw*rad) - (sin(roll*rad)*cos(yaw*rad)))
		       + ACC_Med[0]*(cos(roll*rad)*sin(pitch*rad)*cos(yaw*rad)+   sin(roll*rad)*sin(yaw*rad) );

		ACC_R[1]=  ACC_Med[2]*(sin(roll*rad)*cos(pitch*rad))
	    	       + ACC_Med[1]*(sin(roll*rad)*sin(pitch*rad)*sin(yaw*rad) + (cos(roll*rad)*cos(yaw*rad)))
		       + ACC_Med[0]*(sin(roll*rad)*sin(pitch*rad)*cos(yaw*rad) - (cos(roll*rad)*sin(yaw*rad)));
		       
		ACC_R[0]=  ACC_Med[2]*(-sin(pitch*rad) )
		      +  ACC_Med[1]*(sin(pitch*rad)*sin(yaw*rad) )
		      +  ACC_Med[0]*(cos(pitch*rad)*cos(yaw*rad) );
	
		//ロボット座標系でのGYR
		G_R[2]= G_Global[2]*( cos(roll*rad)*cos(pitch*rad))
	    	       + G_Global[1]*(cos(roll*rad)*sin(pitch*rad)*sin(yaw*rad) - (sin(roll*rad)*cos(yaw*rad)))
		       + G_Global[0]*(cos(roll*rad)*sin(pitch*rad)*cos(yaw*rad)+   sin(roll*rad)*sin(yaw*rad) );

		G_R[1]=  G_Global[2]*(sin(roll*rad)*cos(pitch*rad))
	    	       + G_Global[1]*(sin(roll*rad)*sin(pitch*rad)*sin(yaw*rad) + (cos(roll*rad)*cos(yaw*rad)))
		       + G_Global[0]*(sin(roll*rad)*sin(pitch*rad)*cos(yaw*rad) - (cos(roll*rad)*sin(yaw*rad)));
		       
		G_R[0]=  G_Global[2]*(-sin(pitch*rad) )
		      +  G_Global[1]*(sin(pitch*rad)*sin(yaw*rad) )
		      +  G_Global[0]*(cos(pitch*rad)*cos(yaw*rad) );
		      
	
		ACC_Med_Gcan[0]=(ACC_Med[0]-G_R[0]);
		ACC_Med_Gcan[1]=(ACC_Med[1]-G_R[1]);
		ACC_Med_Gcan[2]=(ACC_Med[2]-G_R[2]);
		ACC_Med_Gcan[0]=Median(7,ACC_Med_Gcan[0],ACC_Med_Gcan_OLD_X);
		ACC_Med_Gcan[1]=Median(7,ACC_Med_Gcan[1],ACC_Med_Gcan_OLD_Y);
		ACC_Med_Gcan[2]=Median(7,ACC_Med_Gcan[2],ACC_Med_Gcan_OLD_Z);
		Velocity_[0]+=(ACC_Med_Gcan[0])*dt;
		Velocity_[1]+=(ACC_Med_Gcan[1])*dt;
		Velocity_[2]+=(ACC_Med_Gcan[2])*dt;	
		yaw_dt  += (sin(roll*rad)*(1/cos(pitch*rad))*GYR_Med[1]*rad*dt+cos(roll*rad)*(1/cos(pitch*rad))*GYR_Med[2]*rad*dt)*degree;
		pitch_dt+= (cos(roll*rad)*GYR_Med[1]*rad*dt-sin(roll*rad)*GYR_Med[2]*rad*dt)*degree;
		roll_dt += (GYR_Med[0]*rad*dt+sin(roll*rad)*tan(pitch*rad)*GYR_Med[1]*rad*dt+cos(roll*rad)*tan(pitch*rad)*GYR_Med[2]*rad*dt)*degree;
 		
	//printf("2:::%f\n",(-cos(roll*rad)*cos(pitch*rad)*sin(yaw*rad) - (sin(roll*rad)*cos(yaw*rad))));
	

	}else{		//メディアンフィルタフラグなし
		//ロボット座標系でのACC
		ACC_R[2]= ACC_Med[2]*( cos(roll*rad)*cos(pitch*rad))
	    	       + ACC_Med[1]*(cos(roll*rad)*sin(pitch*rad)*sin(yaw*rad) - (sin(roll*rad)*cos(yaw*rad)))
		       + ACC_Med[0]*(cos(roll*rad)*sin(pitch*rad)*cos(yaw*rad)+   sin(roll*rad)*sin(yaw*rad) );

		ACC_R[1]=  ACC_Med[2]*(sin(roll*rad)*cos(pitch*rad))
	    	       + ACC_Med[1]*(sin(roll*rad)*sin(pitch*rad)*sin(yaw*rad) + (cos(roll*rad)*cos(yaw*rad)))
		       + ACC_Med[0]*(sin(roll*rad)*sin(pitch*rad)*cos(yaw*rad) - (cos(roll*rad)*sin(yaw*rad)));
		       
		ACC_R[0]=  ACC_Med[2]*(-sin(pitch*rad) )
		      +  ACC_Med[1]*(sin(pitch*rad)*sin(yaw*rad) )
		      +  ACC_Med[0]*(cos(pitch*rad)*cos(yaw*rad) );
	
		//ロボット座標系でのGYR
		G_R[2]= G_Global[2]*( cos(roll*rad)*cos(pitch*rad))
	    	       + G_Global[1]*(cos(roll*rad)*sin(pitch*rad)*sin(yaw*rad) - (sin(roll*rad)*cos(yaw*rad)))
		       + G_Global[0]*(cos(roll*rad)*sin(pitch*rad)*cos(yaw*rad)+   sin(roll*rad)*sin(yaw*rad) );

		G_R[1]=  G_Global[2]*(sin(roll*rad)*cos(pitch*rad))
	    	       + G_Global[1]*(sin(roll*rad)*sin(pitch*rad)*sin(yaw*rad) + (cos(roll*rad)*cos(yaw*rad)))
		       + G_Global[0]*(sin(roll*rad)*sin(pitch*rad)*cos(yaw*rad) - (cos(roll*rad)*sin(yaw*rad)));
		       
		G_R[0]=  G_Global[2]*(-sin(pitch*rad) )
		      +  G_Global[1]*(sin(pitch*rad)*sin(yaw*rad) )
		      +  G_Global[0]*(cos(pitch*rad)*cos(yaw*rad) );
		      
		//ロボット座標系でのMAG
		MAG_R[2]= MAG_Med[2]*( cos(roll*rad)*cos(pitch*rad))
	    	       + MAG_Med[1]*(cos(roll*rad)*sin(pitch*rad)*sin(yaw*rad) - (sin(roll*rad)*cos(yaw*rad)))
		       + MAG_Med[0]*(cos(roll*rad)*sin(pitch*rad)*cos(yaw*rad)+   sin(roll*rad)*sin(yaw*rad) );

		MAG_R[1]=  MAG_Med[2]*(sin(roll*rad)*cos(pitch*rad))
	    	       + MAG_Med[1]*(sin(roll*rad)*sin(pitch*rad)*sin(yaw*rad) + (cos(roll*rad)*cos(yaw*rad)))
		       + MAG_Med[0]*(sin(roll*rad)*sin(pitch*rad)*cos(yaw*rad) - (cos(roll*rad)*sin(yaw*rad)));
		       
		MAG_R[0]=  MAG_Med[2]*(-sin(pitch*rad) )
		      +  MAG_Med[1]*(sin(pitch*rad)*sin(yaw*rad) )
		      +  MAG_Med[0]*(cos(pitch*rad)*cos(yaw*rad) );
	
		ACC_Med_Gcan[0]=(ACC_Med[0]-G_R[0]);
		ACC_Med_Gcan[1]=(ACC_Med[1]-G_R[1]);
		ACC_Med_Gcan[2]=(ACC_Med[2]-G_R[2]);
		Velocity_[0]+=(ACC_Med_Gcan[0])*dt;
		Velocity_[1]+=(ACC_Med_Gcan[1])*dt;
		Velocity_[2]+=(ACC_Med_Gcan[2])*dt;
		yaw_dt  += (sin(roll*rad)*(1/cos(pitch*rad))*GYR_Med[1]*rad*dt+cos(roll*rad)*(1/cos(pitch*rad))*GYR_Med[2]*rad*dt)*degree;
		pitch_dt+= (cos(roll*rad)*GYR_Med[1]*rad*dt-sin(roll*rad)*GYR_Med[2]*rad*dt)*degree;
		roll_dt += (GYR_Med[0]*rad*dt+sin(roll*rad)*tan(pitch*rad)*GYR_Med[1]*rad*dt+cos(roll*rad)*tan(pitch*rad)*GYR_Med[2]*rad*dt)*degree;

	}
	
///////////////////////////////////////////////////////////////////////////////////////////////////////////

	  if(m_Pose_InIn.isNew()){
		 m_Pose_InIn.read();
		/*
		 printf("x:%f y:%f z:%f \n\r"
		 ,m_Pose_In.data.position.x
		 ,m_Pose_In.data.position.y
		 ,m_Pose_In.data.position.z
		 );
		 */
		 Pose_In_Flg=1;
		 //Kal_Cal_XYZ();
	  }else{
		 Pose_In_Flg=0;
	  }

	  if(m_Position_InIn.isNew()){
		 m_Position_InIn.read();
		 Posi_In_Flg=1;
		Kal_Cal_XYZ();
	  }else{
		 Posi_In_Flg=0;
	  }

	  if(m_Orientation_InIn.isNew()){
		 m_Orientation_InIn.read();
		 Orie_In_Flg=1;
		Kal_Cal_XYZ();
	  }else{
		 Orie_In_Flg=0;
	  }
	  
	
  First_Flg=0;//一周回った判定　dtに値あり　　dt ! = 0	大体
	
	
	
	}else{//実際実働部
	//起動後median データが貯まるまで回る用
		cnt++;
	}
	
	
///////////////////////////////////////////////////////////////////////////////////////////////////////////
//    Port書き出し系
///////////////////////////////////////////////////////////////////////////////////////////////////////////
	XYZ[0]=m_Pose_In.data.position.x;
	XYZ[1]=m_Pose_In.data.position.y;
	XYZ[2]=m_Pose_In.data.position.z;
		 
	m_Pose_Out.data.position.x=XYZ[0];
	m_Pose_Out.data.position.y=XYZ[1];
	m_Pose_Out.data.position.z=XYZ[2];
	m_Pose_Out.data.orientation.y=yaw*rad;
	m_Pose_Out.data.orientation.p=pitch*rad;
	m_Pose_Out.data.orientation.r=roll*rad;
	m_Pose_OutOut.write();
	
	m_Position_Out.data.x=XYZ[0];
	m_Position_Out.data.y=XYZ[1];
	m_Position_Out.data.z=XYZ[2];
	m_Position_OutOut.write();
	
	m_Orientation_Out.data.y=yaw*rad;;
	m_Orientation_Out.data.p=pitch*rad;
	m_Orientation_Out.data.r=roll*rad;
	m_Orientation_OutOut.write();
		
	m_Acc.data.ax=ACC_Med[0];
	m_Acc.data.ay=ACC_Med[1];
	m_Acc.data.az=ACC_Med[2];
	m_AccOut.write();
	
	m_Gyro.data.aax=GYR_Med[0];
	m_Gyro.data.aay=GYR_Med[1];
	m_Gyro.data.aaz=GYR_Med[2];
	m_GyroOut.write();
	
	m_Mag.data.x=MAG_Med[0];
	m_Mag.data.y=MAG_Med[1];
	m_Mag.data.z=MAG_Med[2];
	m_MagOut.write();
	
	m_Tmp.data=TMP_Med;
	m_TmpOut.write();
	
	
///////////////////////////////////////////////////////////////////////////////////////////////////////////
//    LOG書き出し系
///////////////////////////////////////////////////////////////////////////////////////////////////////////
	Logout_V<< sum_t 
		<<"\t"<<ACC_Raw[0]<<"\t"<<ACC_Raw[1]<<"\t"<<ACC_Raw[2]
		<<"\t"<<GYR_Raw[0]<<"\t"<<GYR_Raw[1]<<"\t"<<GYR_Raw[2]
		<<"\t"<<MAG_Raw[0]<<"\t"<<MAG_Raw[1]<<"\t"<<MAG_Raw[2]
		<<"\t"
		<<"\t"<<ACC_Med[0]<<"\t"<<ACC_Med[1]<<"\t"<<ACC_Med[2]
		<<"\t"<<GYR_Med[0]<<"\t"<<GYR_Med[1]<<"\t"<<GYR_Med[2]
		<<"\t"<<MAG_Med[0]<<"\t"<<MAG_Med[1]<<"\t"<<MAG_Med[2]
		<<"\t"
		<<"\t"<<ACC_R[0]<<"\t"<<ACC_R[1]<<"\t"<<ACC_R[2]
		<<"\t"<<GYR_R[0]<<"\t"<<GYR_R[1]<<"\t"<<GYR_R[2]
		<<"\t"<<MAG_R[0]<<"\t"<<MAG_R[1]<<"\t"<<MAG_R[2]
		<<"\t"
		<<"\t"<<ACC_Med_Gcan[0]<<"\t"<<ACC_Med_Gcan[1]<<"\t"<<ACC_Med_Gcan[2]
		<<"\t"
		<<"\t"<<Velocity_[0]<<"\t"<<Velocity_[1]<<"\t"<<Velocity_[2]
		<<"\t"
		<<"\t"<<dt<<endl;

	Logout_K<< sum_t 
		<<"\t"<<XYZ[0]<<"\t"<<XYZ[1]<<"\t"<<XYZ[2]
		<<"\t"
		<<"\t"<<yaw<<"\t"<<pitch<<"\t"<<roll
		<<"\t"
		<<"\t"<<XYZ_dt[0]<<"\t"<<XYZ_dt[1]<<"\t"<<XYZ_dt[2]
		<<"\t"
		<<"\t"<<yaw_dt<<"\t"<<pitch_dt<<"\t"<<roll_dt
		<<"\t"<<Direction<<"\t"<<pitch_A*degree<<"\t"<<roll_A*degree
		<<"\t"
		<<"\t"<<Velocity_[0]<<"\t"<<Velocity_[1]<<"\t"<<Velocity_[2]
		<<"\t"
		<<"\t"<<dt<<endl;	
	
///////////////////////////////////////////////////////////////////////////////////////////////////////////
//    時間系
///////////////////////////////////////////////////////////////////////////////////////////////////////////
	gettimeofday(&ti, NULL);
	t0=(double)(ti.tv_sec)+(double)(ti.tv_usec)*0.001*0.001;
        if(First_Flg==1){
        	sum_t=0;
        }else{
	dt=t0-t1;
	sum_t += dt;
	}
	printf("⊿ t:%lf\n\r",dt);
	printf("sum t:%lf\n\r",sum_t);
	printf("\x1b[0J");
	gettimeofday(&ti, NULL);
	t1=(double)(ti.tv_sec)+(double)(ti.tv_usec)*0.001*0.001;
//////////////////////////////////////////////////////////////////////////////////////////////////////////
	
  return RTC::RTC_OK;
}


RTC::ReturnCode_t RTC_NineDOF_Ardiuno::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}


RTC::ReturnCode_t RTC_NineDOF_Ardiuno::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}


RTC::ReturnCode_t RTC_NineDOF_Ardiuno::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}


RTC::ReturnCode_t RTC_NineDOF_Ardiuno::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}


RTC::ReturnCode_t RTC_NineDOF_Ardiuno::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}



RTC::ReturnCode_t RTC_NineDOF_Ardiuno::Connect_serial(){
	//cha PORT=m_SH1_PORT;
        int ret;
        int ERR_FLG=0;
	int p_len;
	char* portname;
	struct termios SERIAL; 
	
//
//m_Portをstringからcharへ
//                     portnameに格納
//m_Portコンフィギュレーションパラメータより
//                     /dev/ttyUSBX X=0,1,2...　　の形
//
	p_len=m_Port.length();
	portname=new char[p_len+1];
	memcpy(portname, m_Port.c_str(), p_len+1);
	
	
	
//Ardiuno　devPATH:portname   O_RDWR:読み書きモードでopen
	Ardiuno=open(portname, O_RDWR);



	std::cout<<portname <<std::endl;	
	
	ret = tcgetattr(Ardiuno, &Ardiuno_ter);
	ret = tcsetattr(Ardiuno, TCSANOW, &Ardiuno_ter);
	if (ret){
			printf("ERR tcsetattr\n\r");
			ERR_FLG=1;
		}
	memset(&SERIAL, 0, sizeof(SERIAL));
//
//c_cflag controlmode
//BAUDRATE->serialの通信速度
//　　　　     　　BAUDRATE コンフィギュレーションパラメータより
//           9600,19200,38400,57600の４択
// CRTSCTS ハードウェアフロー制御ON
// CS8　　　　　8bit Non-parity　stopbit 1
// CLOCAL  Local接続
// CREAD　　　受信文字有効化
//
	        if(m_BAUDRATE.compare("B9600")==0){
	        	SERIAL.c_cflag =(B9600 | CRTSCTS | CS8 | CLOCAL | CREAD);
		//        std::cout<<"Baudrate B9600" <<std::endl;	
	        }else if(m_BAUDRATE.compare("B19200")==0){
	        	SERIAL.c_cflag =(B19200 | CRTSCTS | CS8 | CLOCAL | CREAD);
		//        std::cout<<"Baudrate B19200" <<std::endl;	
	        }else if(m_BAUDRATE.compare("B38400")==0){
	        	SERIAL.c_cflag =(B38400 | CRTSCTS | CS8 | CLOCAL | CREAD);
		 //       std::cout<<"Baudrate B38400" <<std::endl;	
	        }else if(m_BAUDRATE.compare("B57600")==0){
	        	SERIAL.c_cflag =(B57600 | CRTSCTS | CS8 | CLOCAL | CREAD);
		 //       std::cout<<"Baudrate B57600" <<std::endl;	
	        }else if(m_BAUDRATE.compare("B115200")==0){
	        	SERIAL.c_cflag =(B115200 | CRTSCTS | CS8 | CLOCAL | CREAD);
		 //       std::cout<<"Baudrate B57600" <<std::endl;	
	        }else{
		        std::cout<<"Baudrate ERR" <<std::endl;	
	        }	       
//
//c_iflag   inputmode
// IGNPAR  Parityエラー無視
// ICRNL　　　入力のCR→NLに置き換え
//
		SERIAL.c_iflag = (IGNPAR | ICRNL);//IGNBRK | IGNPAR;if (ret < 0)
		//printf("ERR");
	     //   cfsetispeed(&tio, m_BAUDRATE1);
		ret = tcflush(Ardiuno, TCIFLUSH);
	
		if (ret < 0){
			printf("ERR tcflush\n\r");
			ERR_FLG=1;
		}
		//   .// exitfail_errno("tcflush");

	 
		ret = tcsetattr(Ardiuno, TCSANOW, &SERIAL);
		if (ret){
			printf("ERR tcsetattr\n\r");
			ERR_FLG=1;
		}
		
		if(ERR_FLG!=1){
			printf("OK!\n\r");
			SERIAL_OK_FLG[0]=1;
		}
		else{
			SERIAL_OK_FLG[0]=0;
		}
		
        
                
  return RTC::RTC_OK;
}






//RTC::ReturnCode_t RTC_NineDOF_Ardiuno::Connect_serial(){

RTC::ReturnCode_t RTC_NineDOF_Ardiuno::Kal_Cal_YPR()
{
///////////////////////////////////////////////////////////////////////////////////////////////////////////
//    カルマンフィルタ
///////////////////////////////////////////////////////////////////////////////////////////////////////////
	printf("Direction:%f\n",Direction);
	if (m_Kalman_ON_OFF == 1) {
			//加速度・地磁気を元に観測値生成
			if (m_Orientation_FIX==0){
				cal_YPR(ACC_Med,GYR_Med,MAG_Med,yaw_M,pitch_A*degree,roll_A*degree);
				//cal_XYZ(ACC_,double X,double Y,double Z);
			}	
			//以下　観測値外部入力
			//入力　PoseもしくはOrientation
			//どちらも入力がある場合Poseのほうが優先度↑
			//	外部入力フラグあり（！＝０）。入力なき場合は通常の地磁気・加速度による観測値
			else if(m_Orientation_FIX==1){

			//yaw のみ

				if(Pose_In_Flg == 1){
					cal_YPR(ACC_Med,GYR_Med,MAG_Med
					, m_Pose_In.data.orientation.y*degree
					,pitch_A*degree
					,roll_A*degree);
		
				}else if(Orie_In_Flg==1){
					cal_YPR(ACC_Med,GYR_Med,MAG_Med
					, m_Orientation_In.data.y*degree
					,pitch_A*degree
					,roll_A*degree);
				}else{
					cal_YPR(ACC_Med,GYR_Med,MAG_Med,yaw_M,pitch_A*degree,roll_A*degree);
				}

			}else if(m_Orientation_FIX==2){

			//Pitch のみ
		
				if(Pose_In_Flg == 1){
					cal_YPR(ACC_Med,GYR_Med,MAG_Med
					, yaw_M
					,m_Pose_In.data.orientation.p*degree
					,roll_A*degree);
		
				}else if(Orie_In_Flg==1){
					cal_YPR(ACC_Med,GYR_Med,MAG_Med
					, yaw_M
					,m_Orientation_In.data.p*degree
					,roll_A*degree);
				}else{
					cal_YPR(ACC_Med,GYR_Med,MAG_Med,yaw_M,pitch_A*degree,roll_A*degree);
				}


			}else if(m_Orientation_FIX==3){
		
			//Roll のみ
	
				if(Pose_In_Flg == 1){
					cal_YPR(ACC_Med,GYR_Med,MAG_Med
					, yaw_M
					,pitch_A*degree
					,m_Pose_In.data.orientation.r*degree
					);
		
				}else if(Orie_In_Flg==1){
					cal_YPR(ACC_Med,GYR_Med,MAG_Med
					, yaw_M
					,pitch_A*degree
					,m_Orientation_In.data.r*degree
					);
				}else{
					cal_YPR(ACC_Med,GYR_Med,MAG_Med,yaw_M,pitch_A*degree,roll_A*degree);
				}

			}else if(m_Orientation_FIX==4){
	
			//yaw Pitch

				if(Pose_In_Flg == 1){
					cal_YPR(ACC_Med,GYR_Med,MAG_Med
					, m_Pose_In.data.orientation.y*degree
					,m_Pose_In.data.orientation.p*degree
					,roll_A*degree);
		
				}else if(Orie_In_Flg==1){
					cal_YPR(ACC_Med,GYR_Med,MAG_Med
					, m_Orientation_In.data.y*degree
					, m_Orientation_In.data.p*degree
					,roll_A*degree);
				}else{
					cal_YPR(ACC_Med,GYR_Med,MAG_Med,yaw_M,pitch_A*degree,roll_A*degree);
				}

			}else if(m_Orientation_FIX==5){
	
			//yaw roll

				if(Pose_In_Flg == 1){
					cal_YPR(ACC_Med,GYR_Med,MAG_Med
					, m_Pose_In.data.orientation.y*degree
					,pitch_A*degree
					, m_Pose_In.data.orientation.r*degree);
		
				}else if(Orie_In_Flg==1){
					cal_YPR(ACC_Med,GYR_Med,MAG_Med
					, m_Orientation_In.data.y*degree
					,pitch_A*degree
					, m_Orientation_In.data.r*degree);
				}else{
					cal_YPR(ACC_Med,GYR_Med,MAG_Med,yaw_M,pitch_A*degree,roll_A*degree);
				}

			}else if(m_Orientation_FIX==6){
	
			//Pitch roll

				if(Pose_In_Flg == 1){
					cal_YPR(ACC_Med,GYR_Med,MAG_Med
					, yaw_M
					,m_Pose_In.data.orientation.p*degree
					,m_Pose_In.data.orientation.r*degree);
		
				}else if(Orie_In_Flg==1){
					cal_YPR(ACC_Med,GYR_Med,MAG_Med
					, yaw_M
					, m_Orientation_In.data.p*degree
					, m_Orientation_In.data.r*degree);
				}else{
					cal_YPR(ACC_Med,GYR_Med,MAG_Med,yaw_M,pitch_A*degree,roll_A*degree);
				}

			}else if(m_Orientation_FIX==7){
	
			//Yaw Pitch roll

				if(Pose_In_Flg == 1){
					cal_YPR(ACC_Med,GYR_Med,MAG_Med
					,m_Pose_In.data.orientation.y*degree
					,m_Pose_In.data.orientation.p*degree
					,m_Pose_In.data.orientation.r*degree);
		
				}else if(Orie_In_Flg==1){
					cal_YPR(ACC_Med,GYR_Med,MAG_Med
					, m_Orientation_In.data.y*degree
					, m_Orientation_In.data.p*degree
					, m_Orientation_In.data.r*degree);
				}else{
					cal_YPR(ACC_Med,GYR_Med,MAG_Med,yaw_M,pitch_A*degree,roll_A*degree);
				}

			}else{
					cal_YPR(ACC_Med,GYR_Med,MAG_Med,yaw_M,pitch_A*degree,roll_A*degree);
			}
			
		
	}else{
///////////////////////////////////////////////////////////////////////////////////////////////////////////
//    カルマンフィルタなし
///////////////////////////////////////////////////////////////////////////////////////////////////////////
		if(Ori_dt_Flg==0){//初回ループdtがない　
			roll += roll_A;
			pitch += pitch_A;
			yaw  += yaw_M;
		}else{
		
			yaw   += (sin(roll*rad)*(1/cos(pitch*rad))*GYR_Med[1]*rad*dt+cos(roll*rad)*(1/cos(pitch*rad))*GYR_Med[2]*rad*dt)*degree;
			pitch += (cos(roll*rad)*GYR_Med[1]*rad*dt-sin(roll*rad)*GYR_Med[2]*rad*dt)*degree;
			roll   += (GYR_Med[0]*rad*dt+sin(roll*rad)*tan(pitch*rad)*GYR_Med[1]*rad*dt+cos(roll*rad)*tan(pitch*rad)*GYR_Med[2]*rad*dt)*degree;
	 		
		
		}


	}
  return RTC::RTC_OK;
}


//RTC::ReturnCode_t RTC_NineDOF_Ardiuno::Connect_serial(){

RTC::ReturnCode_t RTC_NineDOF_Ardiuno::Kal_Cal_XYZ(){

  
	
///////////////////////////////////////////////////////////////////////////////////////////////////////////
//    G-Cancel　より　Velocity -> XYZ
///////////////////////////////////////////////////////////////////////////////////////////////////////////
	
	if (m_Kalman_ON_OFF == 1) {
			if (m_Position_FIX==0){
				if(Pose_In_Flg == 1){
					cal_XYZ(Velocity_
					,m_Pose_In.data.position.x
					,m_Pose_In.data.position.y
					,m_Pose_In.data.position.z);
		
				}else if(Orie_In_Flg==1){
					cal_XYZ(Velocity_
					, m_Position_In.data.x
					, m_Position_In.data.y
					, m_Position_In.data.z);
				}else{
				
					XYZ_dt[0]+=Velocity_[0]*dt;
					XYZ_dt[1]+=Velocity_[1]*dt;
					XYZ_dt[2]+=Velocity_[2]*dt;
					XYZ[0]=XYZ_dt[0];
					XYZ[1]=XYZ_dt[1];
					XYZ[2]=XYZ_dt[2];
					//cal_YPR(ACC_Med,GYR_Med,MAG_Med,yaw_M,pitch_A*degree,roll_A*degree);
				}
			}	
	}
	
	
	
  return RTC::RTC_OK;
}










void mat_mul(double A[3][3],double B[3][3],double C[3][3]){
	int i=0,j=0,k=0;
	for(i=0;i<3;i++){
		for(j=0;j<3;j++){
			C[i][j]=0;
		}
	}
	for(i=0;i<3;i++){
		for(j=0;j<3;j++){
			for(k=0;k<3;k++){
				C[i][j] += A[i][k]*B[k][i];
			}
			
		}
	}

 }
 void mat_pul(double A[3][3],double B[3][3],double C[3][3]){
	int i=0,j=0,k=0;
	for(i=0;i<3;i++){
		for(j=0;j<3;j++){
			C[i][j]=0;
		}
	}
	for(i=0;i<3;i++){
		for(j=0;j<3;j++){
				C[i][j] = A[i][j]+B[i][j];
						
		}
	}

 }
 void mat_m(double A[3][3],double B[3][3],double C[3][3]){
	int i=0,j=0,k=0;
	for(i=0;i<3;i++){
		for(j=0;j<3;j++){
			C[i][j]=0;
		}
	}
	
	for(i=0;i<3;i++){
		for(j=0;j<3;j++){
				C[i][j] = A[i][j]-B[i][j];
						
		}
	}

 }
 
 void T_33(double A[3][3],double B[3][3]){
 	int i=0,j=0,k=0;
	for(i=0;i<3;i++){
		for(j=0;j<3;j++){
				B[i][j] = A[j][i];
						
		}
	}
 }
 void inv_33(double A[3][3],double B[3][3]){
	double detA=0;
	int i=0,j=0,k=0;
	detA=
	 A[0][0]*A[1][1]*A[2][2]
	+A[1][0]*A[2][1]*A[1][3]
	+A[2][0]*A[0][1]*A[1][2]
	-A[0][0]*A[2][1]*A[1][2]
	-A[2][0]*A[1][1]*A[0][2]
	-A[1][0]*A[0][1]*A[2][2];	
	if(detA!=0){
	
		B[0][0]=A[1][1]*A[2][2]-A[1][2]*A[2][1];
		B[1][0]=A[1][2]*A[2][0]-A[1][0]*A[2][2];
		B[2][0]=A[1][0]*A[2][1]-A[1][1]*A[2][0];
		
		B[0][1]=A[0][2]*A[2][1]-A[0][1]*A[2][2];
		B[1][1]=A[0][0]*A[2][2]-A[0][2]*A[2][0];
		B[2][1]=A[0][1]*A[2][0]-A[0][0]*A[2][1];
		
		B[0][2]=A[0][1]*A[1][2]-A[0][2]*A[1][1];
		B[1][2]=A[0][2]*A[1][0]-A[0][0]*A[1][2];
		B[2][2]=A[0][0]*A[1][1]-A[0][1]*A[1][0];
		
		B[0][0]/=detA;
		B[1][0]/=detA;
		B[2][0]/=detA;
		
		B[0][1]/=detA;
		B[1][1]/=detA;
		B[2][1]/=detA;
		
		B[0][2]/=detA;
		B[1][2]/=detA;
		B[2][2]/=detA;
	
	}
	
}


double Median(int Len,double new_data,double* data){
int i,j;
double buf=0;
double data_out[256];
//Len=sizeof(data);
     data_out[0]=new_data;
     data[0]=new_data;
     for (i = 1; i < Len; i++)data_out[i]=data[i]; 

   //  printf("==============\n");
   // for (i = 0; i < Len; i++) printf("data_o[%d]:%lf\n",i,data_out[i]);       
   // printf("==============\n");
     for (j = 0; j < Len; j++) {
        for (i = 0; i < Len; i++) {
            if (data_out[i+1] < data_out[i]) {
                buf = data_out[i+1];
                data_out[i+1] = data_out[i];
                data_out[i] = buf;
              }
         }
      }
     //printf("==============\n");
     //for (i = 0; i < Len; i++) printf("data[%d]:%lf\n",i,data_out[i]);       
     //printf("==============\n");
     //printf("%d\n",i);
   //  int n=(Len-1);
     for (i =7; i > 0; i=i-1){
     	data[i]=data[i-1];
     //	printf("[%d]:%lf=[%d]:%lf\n",i,data[i],i-1,data[i-1]);   
     }        
     //printf("==============\n");
     //for (i = 0; i < Len; i++) printf("data[%d]:%lf\n",i,data[i]);       
     //printf("==============\n");
     
return data_out[(Len-1)/2];
}
double Direction_Estimate(double ACC_Med[3],double MAG_Med[3],double r,double p,double OFFSET)
{
  
  double Med1_x;
  double Med1_y;
  double Med1_z;
  double Med1;
  
  double Med2_x;
  double Med2_y;
  double Med2_z;
  double Med2;
  //double Direction;
  
  double Gx;
  double Gy;
  double Gz;
  
  
  double mx;
  double my;
  double mz;
  //
  //magはaxと座標系が異なる
  //	axに合わせたxyzへ
  //	ax:my ay:mx az:-mz
  //　　　　Direction 
  //        0　北
  mx=MAG_Med[1];
  my=MAG_Med[0];
  mz=MAG_Med[2];
  
  Med1_x=my*ACC_Med[2]-mz*ACC_Med[1];
  Med1_y=mz*ACC_Med[0]-mx*ACC_Med[2];
  Med1_z=mx*ACC_Med[1]-my*ACC_Med[0];
  Med1=sqrt(Med1_x*Med1_x+Med1_y*Med1_y+Med1_z*Med1_z);
  
  Med2_x=ACC_Med[1]*Med1_z-ACC_Med[2]*Med1_y;
  Med2_y=ACC_Med[2]*Med1_x-ACC_Med[0]*Med1_z;
  Med2_z=ACC_Med[0]*Med1_y-ACC_Med[1]*Med1_x;
  Med2=sqrt(Med2_x*Med2_x+Med2_y*Med2_y+Med2_z*Med2_z);
  /*
  Gx=m__acc.data.ax;
  Gy=m__acc.data.ay;
  Gz=m__acc.data.az;
  */
  
  printf("p1%f  p2%f",Med1,Med2);
  if(Med1==0||Med2==0){
  	Direction=0;
  }else{
  	Direction=atan2(Med1_x/Med1,Med2_x/Med2)*degree;
  }
/*
	Med1_y=mz*(sin(-roll*rad)*cos(-pitch*rad))
       + my*(sin(-roll*rad)*sin(-pitch*rad)*sin(0*rad) + (cos(-roll*rad)*cos(0*rad)))
       + mx*(sin(-roll*rad)*sin(-pitch*rad)*cos(0*rad) - (cos(-roll*rad)*sin(0*rad)));
		       
	Med1_x= mz*(-sin(-pitch*rad) )
      +  my*(sin(-pitch*rad)*sin(0*rad) )
      +  mx*(cos(-pitch*rad)*cos(0*rad) );

*/
	Med1_y= my*((cos(pitch*rad)))
       + sin(pitch*rad)*( mx*sin(roll*rad) - mz*cos(roll*rad) ) ;
		       
	Med1_x= mz*(sin(roll*rad) )
      +  mx*(cos(roll*rad));
      
  //Direction=atan2(-Med1_y,Med1_x)*degree;
  Direction=-atan2(Med1_y,Med1_x)*degree;
  if( (Direction-Direction_OLD) > 300)
      Dir_cnt++;
  if( (Direction-Direction_OLD) < -300)
      Dir_cnt--;
  Direction_OLD=Direction;
  Direction=atan2(-Med1_y,Med1_x)*degree-360*Dir_cnt + (OFFSET);
  printf("\x1b[34m");
  printf("Direction:%f\n",Direction);
  return Direction;
}

//
//カルマン　姿勢用　のみ
//
//void cal_YPR(double* ACC_,double* GYR_,double* MAG_,double Y_0,double Y_1,double Y_2){

RTC::ReturnCode_t RTC_NineDOF_Ardiuno::cal_YPR(double* ACC_,double* GYR_,double* MAG_,double Y_0,double Y_1,double Y_2){
 	double tmp=0;
 	double TMP[3][3];
 	double TMP2[3][3];
 	double A[3][3]={ {1 ,0 ,0},
		 	 {0 ,1 ,0},  
 	 		 {0 ,0 ,1} };
 	double AT[3][3];
 	double C[3][3]={ {1 ,0 ,0},
		 	 {0 ,1 ,0},  
 	 		 {0 ,0 ,1} };
 	double CT[3][3];
 	double P0[3][3]={{0 ,0 ,0},
		 	 {0 ,0 ,0},  
 	 		 {0 ,0 ,0} };
 	double Q[3][3]={{ 2 ,0 ,0},
		 	 {0 ,2 ,0},  
 	 		 {0 ,0,2} };
 	double K_mul_INV[3][3];
 	double R[3][3]={{5,0, 0},
		 	 {0 ,5 ,0},  
 	 		 {0 ,0 ,5} };
 	double I[3][3]={{1 ,0 ,0},
		 	 {0 ,1 ,0},  
 	 		 {0 ,0 ,1} };
 	double r_d=0,p_d=0,y_d=0;
 	
 	if(First_Flg==1)
 	{
	roll= roll_A;
 	pitch= pitch_A;
 	yaw =Direction;
 	
	}
 	else{
		  yaw  *=rad;
	 	  pitch*=rad;
 		  roll *=rad;
 		
 		y_d=sin(roll)*(1/cos(pitch))*GYR_[1]*rad*dt+cos(roll)*(1/cos(pitch))*GYR_[2]*rad*dt;
 		p_d=cos(roll)*GYR_[1]*rad*dt-sin(roll)*GYR_[2]*rad*dt;
 		r_d=GYR_[0]*rad*dt+sin(roll)*tan(pitch)*GYR_[1]*rad*dt+cos(roll)*tan(pitch)*GYR_[2]*rad*dt;

	  	 yaw  *=degree;
	 	 pitch*=degree;
 		 roll *=degree;
	 	
 	}
 	
 	
 	
	//printf("\n\r%f\n\r",y_d*degree);
         //////////////////////////////////
	 //   X-
	 ////////////////////////////////// 
	 if(First_Flg==1){		
		  YPR_[0]=yaw_dt;
		  YPR_[1]=pitch_dt;
		  YPR_[2]=roll_dt;
		  YPR_O[0]=y_d*degree;
		  YPR_O[1]=p_d*degree;
		  YPR_O[2]=r_d*degree;
		  P0[0][0]=YPR_[0];
		  P0[1][1]=YPR_[1];
		  P0[2][2]=YPR_[2];
	  }else{
		  YPR_[0]=YPR[0]+y_d*degree;//yaw+y_d*degree;
		  YPR_[1]=YPR[1]+p_d*degree;//pitch+p_d*degree;
		  YPR_[2]=YPR[2]+r_d*degree;//roll+r_d*degree;
		  YPR_O[0]=yaw+y_d*degree;
		  YPR_O[1]=pitch+p_d*degree;
		  YPR_O[2]=roll+r_d*degree;
		  
	  }
	  T_33(A,AT);
         //////////////////////////////////
	 //   APAT
	 //////////////////////////////////	
	  if(First_Flg==1){	  
		  mat_mul(A,P0,TMP);
		  mat_mul(TMP,AT,TMP2);
	  }else{
		  mat_mul(A,P_YPR,TMP);
		  mat_mul(TMP,AT,TMP2);
	  }
//	  P_YPR_=TMP+Q;
	  mat_pul(TMP2,Q,P_YPR_);
	  
	  printf("\n\r");
	  T_33(C,CT);
	  
	  mat_mul(C,P_YPR_,TMP);
	  mat_mul(TMP,CT,TMP2);
	  mat_pul(TMP2,R,TMP);
	  inv_33(TMP,K_mul_INV);
	  
	  mat_mul(P_YPR_,CT,TMP2);
	  mat_mul(TMP2,K_mul_INV,K);
	 
	  
	  YPR[0]=YPR_[0]+K[0][0]*(Y_0-YPR_[0]);
//	  +K[0][1]*(ANG_P[1]-YPR_[1])
//	  +K[0][2]*(ANG_P[0]-YPR_[2]);
	  
	  YPR[1]=YPR_[1]//+K[1][0]*(ANG_P[2]-YPR_[0])
	  +K[1][1]*(/*ANG_P[1]*/Y_1-YPR_[1]);
//	  +K[1][2]*(ANG_P[0]-YPR_[2]);
	  
	  YPR[2]=YPR_[2]//+K[2][0]*(ANG_P[2]-YPR_[0])
//	  +K[2][1]*(ANG_P[1]-YPR_[1])
	  +K[2][2]*(/*ANG_P[0]*/Y_2-YPR_[2]);

	  
	  tmp=ACC_[1]*ACC_[1]+ACC_[2]*ACC_[2];
	  
	  
	  mat_mul(K,C,TMP);
	  mat_m(I,TMP,TMP2);
	  mat_mul(TMP2,P_YPR_,P_YPR);
	  
	  yaw=YPR[0];
 	  pitch=YPR[1];
 	  roll=YPR[2];
	printf("\n\rM_Y:%f K_Y:%f\n\r",Direction,yaw);
	printf("A_P:%f K_P:%f \n\r",pitch_A,pitch);
	printf("A_R:%f K_R%f \n\r",roll_A, roll);
  return RTC::RTC_OK;
 
 }




//
//カルマン　位置用　のみ
//

void cal_XYZ(double* ACC_,double X,double Y,double Z){
 	double tmp=0;
 	double TMP[3][3];
 	double TMP2[3][3];
 	double A[3][3]={ {1 ,0 ,0},
		 	 {0 ,1 ,0},  
 	 		 {0 ,0 ,1} };
 	double AT[3][3];
 	double C[3][3]={ {1 ,0 ,0},
		 	 {0 ,1 ,0},  
 	 		 {0 ,0 ,1} };
 	double CT[3][3];
 	double P0[3][3]={{0 ,0 ,0},
		 	 {0 ,0 ,0},  
 	 		 {0 ,0 ,0} };
 	double Q[3][3]={{ 2 ,0 ,0},
		 	 {0 ,2 ,0},  
 	 		 {0 ,0,2} };
 	double K_mul_INV[3][3];
 	double R[3][3]={{1,0, 0},
		 	 {0 ,1 ,0},  
 	 		 {0 ,0 ,1} };
 	double I[3][3]={{1 ,0 ,0},
		 	 {0 ,1 ,0},  
 	 		 {0 ,0 ,1} };
 	double x_d=0,y_d=0,z_d=0;
 	
 	if(First_Flg==1)
 	{
 //	roll= roll_A;
 //	pitch=pitch_A;
 //	yaw =Direction;
 	}
 	else{
// 		XYZ[0]+=Velocity_[0]*dt;
//		XYZ[1]+=Velocity_[1]*dt;
//		XYZ[2]+=Velocity_[2]*dt;
		
 		x_d=ACC_[0]*dt;;//cos(roll)*GYR_[1]*rad*dt-sin(roll)*GYR_[2]*rad*dt;
 		y_d=ACC_[1]*dt;;//sin(roll)*(1/cos(pitch))*GYR_[1]*rad*dt+cos(roll)*(1/cos(pitch))*GYR_[2]*rad*dt;
 		z_d=ACC_[2]*dt;;//GYR_[0]*rad*dt+sin(roll)*tan(pitch)*GYR_[1]*rad*dt+cos(roll)*tan(pitch)*GYR_[2]*rad*dt;
 		
	 	
 	}
 	
 	
 	
	//printf("\n\r%f\n\r",y_d*degree);
         //////////////////////////////////
	 //   X-
	 ////////////////////////////////// 
	 if(First_Flg==1){		
		  XYZ[0]=0;
		  XYZ[1]=0;
		  XYZ[2]=0;
	  }else{
		  XYZ[0]=XYZ[0]+x_d;
		  XYZ[1]=XYZ[1]+y_d;
		  XYZ[2]=XYZ[2]+z_d;
		  
	  }
	  T_33(A,AT);
         //////////////////////////////////
	 //   APAT
	 //////////////////////////////////	
	  if(First_Flg==1){	  
		  mat_mul(A,P0,TMP);
		  mat_mul(TMP,AT,TMP2);
	  }else{
		  mat_mul(A,P_XYZ,TMP);
		  mat_mul(TMP,AT,TMP2);
	  }
//	  P_XYZ_=TMP+Q;
	  mat_pul(TMP2,Q,P_XYZ_);
	  
	  printf("\n\r");
	  T_33(C,CT);
	  
	  mat_mul(C,P_XYZ_,TMP);
	  mat_mul(TMP,CT,TMP2);
	  mat_pul(TMP2,R,TMP);
	  inv_33(TMP,K_mul_INV);
	  
	  mat_mul(P_XYZ_,CT,TMP2);
	  mat_mul(TMP2,K_mul_INV,K_XYZ);
	 
	  
	  XYZ[0]=XYZ_[0]+K_XYZ[0][0]*(X-XYZ_[0]);
//	  +K[0][1]*(ANG_P[1]-XYZ_[1])
//	  +K[0][2]*(ANG_P[0]-XYZ_[2]);
	  
	  XYZ[1]=XYZ_[1]//+K[1][0]*(ANG_P[2]-XYZ_[0])
	  +K_XYZ[1][1]*(/*ANG_P[1]*/Y-XYZ_[1]);
//	  +K[1][2]*(ANG_P[0]-XYZ_[2]);
	  
	  XYZ[2]=XYZ_[2]//+K[2][0]*(ANG_P[2]-XYZ_[0])
//	  +K[2][1]*(ANG_P[1]-XYZ_[1])
	  +K_XYZ[2][2]*(/*ANG_P[0]*/Z-XYZ_[2]);

	  
	  
	  mat_mul(K,C,TMP);
	  mat_m(I,TMP,TMP2);
	  mat_mul(TMP2,P_XYZ_,P_XYZ);
	  
	  XYZ[0]=XYZ[0];
 	  XYZ[1]=XYZ[1];
 	  XYZ[2]=XYZ[2];
	  printf("\n\rX:%f\n\r",XYZ[0]);
	  printf("Y:%f \n\r",XYZ[1]);
	  printf("Z%f \n\r",XYZ[2]);
 
 }

void LOG_init(){
	Logout_V<<"TIME"
		<<"\t"<<"ACC_X_Raw"<<"\t"<<"ACC_Y_Raw"<<"\t"<<"ACC_Z_Raw"
		<<"\t"<<"GYR_X_Raw"<<"\t"<<"GYR_Y_Raw"<<"\t"<<"GYR_Z_Raw"
		<<"\t"<<"MAG_X_Raw"<<"\t"<<"MAG_Y_Raw"<<"\t"<<"MAG_Z_Raw"
		<<"\t"
		<<"\t"<<"ACC_X_Med"<<"\t"<<"ACC_Y_Med"<<"\t"<<"ACC_Z_Med"
		<<"\t"<<"GYR_X_Med"<<"\t"<<"GYR_Y_Med"<<"\t"<<"GYR_Z_Med"
		<<"\t"<<"MAG_X_Med"<<"\t"<<"MAG_Y_Med"<<"\t"<<"MAG_Z_Med"
		<<"\t"
		<<"\t"<<"ACC_X_Local"<<"\t"<<"ACC_Y_Local"<<"\t"<<"ACC_Z_Local"
		<<"\t"<<"GYR_X_Local"<<"\t"<<"GYR_Y_Local"<<"\t"<<"GYR_Z_Local"
		<<"\t"<<"MAG_X_Local"<<"\t"<<"MAG_Y_Local"<<"\t"<<"MAG_Z_Local"
		<<"\t"
		<<"\t"<<"ACC_X_G-Cancel"<<"\t"<<"ACC_Y_G-Cancel"<<"\t"<<"ACC_Z_G-Cancel"
		<<"\t"
		<<"\t"<<"Velocity_X"<<"\t"<<"Velocity_Y"<<"\t"<<"Velocity_Z"
		<<"\t"
		<<"\t"<<"dt"<<endl;

	Logout_K<< sum_t 
		<<"\t"<<"Position_X"<<"\t"<<"Position_Y"<<"\t"<<"Position_Z"
		<<"\t"
		<<"\t"<<"Yaw"<<"\t"<<"Pitch"<<"\t"<<"Roll"
		<<"\t"
		<<"\t"<<"Position_X_dt"<<"\t"<<"Position_Y_dt"<<"\t"<<"Position_Z_dt"
		<<"\t"
		<<"\t"<<"Yaw_dt"<<"\t"<<"Pitch_dt"<<"\t"<<"Roll_dt"
		<<"\t"<<"Yaw_Observed"<<"\t"<<"Pitch_Observed"<<"\t"<<"Roll__Observed"
		<<"\t"
		<<"\t"<<"Velocity_X"<<"\t"<<"Velocity_Y"<<"\t"<<"Velocity_Z"
		<<"\t"
		<<"\t"<<"dt"<<endl;	
	
}


extern "C"
{
 
  void RTC_NineDOF_ArdiunoInit(RTC::Manager* manager)
  {
    coil::Properties profile(rtc_ninedof_ardiuno_spec);
    manager->registerFactory(profile,
                             RTC::Create<RTC_NineDOF_Ardiuno>,
                             RTC::Delete<RTC_NineDOF_Ardiuno>);
  }
  
};


